import numpy as np
import sys
from matplotlib.textpath import TextPath
from matplotlib.patches import PathPatch, FancyBboxPatch
from matplotlib.patches import Path
from pyquaternion import Quaternion
from os.path import exists

def StrToBool(s):
	if isinstance(s, bool):
		return s
	if s.lower() in ('yes', 'true', 't', 'y', '1'):
		return True
	elif s.lower() in ('no', 'false', 'f', 'n', '0'):
		return False
	else:
		raise argparse.ArgumentTypeError('Boolean value expected.')

def StrToNumpy(s):
	if not s:
		return np.array([])
	else:
		l = list(map(int,s.split(':')))
		return np.array(l) if len(l)==1 else  np.arange(l[0],l[1]) if len(l)==2 else np.arange(l[0],l[1],l[2])

def Arrow(ax,Xa,Xb,**opt):
	def get_key(key,default_value):
		return default_value if key not in opt else opt[key]
	arrowstyle = get_key('arrowstyle','-')
	xycoords = get_key('xycoords','data')
	if (any(elem in arrowstyle for elem in '<>')):
		head_length = get_key('head_length','0.8');
		head_width  = get_key('head_width' ,'0.4');
		arrowstyle+=',head_length = ' + str(head_length) + ',head_width = ' + str(head_width)
	return ax.annotate(get_key('text',''),
	                   xytext = (Xa[0],Xa[1]),
	                   xy = (Xb[0],Xb[1]),
	                   xycoords = xycoords,
	                   textcoords = xycoords,
	                   zorder = get_key('zorder',-1),
	                   annotation_clip = get_key('annotation_clip',False),
	                   ha='center',
	                   va='center',
	                   arrowprops = dict(arrowstyle = arrowstyle,
	                                     shrinkA = get_key('shrinkA',0),
	                                     shrinkB = get_key('shrinkB',0),
	                                     linestyle = get_key('linestyle','-'),
	                                     linewidth = get_key('linewidth',1),
	                                     facecolor = get_key('color','k'),
	                                     edgecolor = get_key('color','k'),
	                                     joinstyle = get_key('joinstyle','miter'),
	                                     clip_path = get_key('clip',None)))

def TransformFigToAx(ax,x_fig,y_fig):
	bbox = ax.get_position()
	xlim = ax.get_xlim()
	ylim = ax.get_ylim()
	alphax = (x_fig-bbox.x0) / (bbox.x1-bbox.x0)
	alphay = (y_fig-bbox.y0) / (bbox.y1-bbox.y0)
	x_ax = xlim[0] + alphax*(xlim[1]-xlim[0]);
	y_ax = ylim[0] + alphay*(ylim[1]-ylim[0]);
	return x_ax,y_ax

def TransformAxToFig(ax,x_ax,y_ax):
	bbox = ax.get_position()
	xlim = ax.get_xlim()
	ylim = ax.get_ylim()
	alphax = (x_ax-xlim[0])/(xlim[1]-xlim[0])
	alphay = (y_ax-ylim[0])/(ylim[1]-ylim[0])
	xnew = bbox.x0 + alphax * (bbox.x1-bbox.x0)
	ynew = bbox.y0 + alphay * (bbox.y1-bbox.y0)
	return xnew,ynew

def TransformAxToData(ax,x_ax,y_ax):
	xlim = ax.get_xlim(); x_ax=xlim[0]+x_ax*(xlim[1]-xlim[0])
	ylim = ax.get_ylim(); y_ax=ylim[0]+y_ax*(ylim[1]-ylim[0])
	return x_ax,y_ax

def TextBox(ax,X,text,**opt):
	def get_key(key,default_value=-1):
		return default_value if key not in opt else opt[key]
	tp = TextPath((0,0),
	              text,
	              size = get_key('fontsize',10),
	              usetex = get_key('usetex',True))
	bb = tp.get_extents()
	xm,ym = (bb.x0 + bb.x1)/2,(bb.y0 + bb.y1)/2
	if (get_key('xycoords') == 'axes fraction'):
		X[0],X[1] = TransformAxToData(ax,X[0],X[1])
	tp = Path(tp.vertices.copy(),codes=tp.codes.copy()) #tp = tp.deepcopy()
	tp.vertices[:,0] = np.add(tp.vertices[:,0],-xm+X[0])
	tp.vertices[:,1] = np.add(tp.vertices[:,1],-ym+X[1])
	pp = PathPatch(tp,
	               facecolor = get_key('textcolor','black'),
	               edgecolor = 'None',
	               #linewidth = None,
	               clip_on = get_key('clip_on',False),
	               zorder = get_key('zorder',-1))
	bb = tp.get_extents()
	rect = FancyBboxPatch((bb.x0,bb.y0),
	                      bb.width,
	                      bb.height,
	                      boxstyle = 'round,pad={},rounding_size={}'.format(get_key('pad',0),get_key('rounding_size',0)),
	                      facecolor = get_key('facecolor','white'),
	                      edgecolor = get_key('edgecolor','black'),
	                      linewidth = get_key('linewidth',1),
	                      clip_on = get_key('clip_on',False),
	                      alpha = get_key('alpha_box',1),
	                      zorder = get_key('zorder',-1))
	ax.add_artist(rect)
	ax.add_artist(pp)

#
#  PETSC POSTPROCESSING STUFF
#

def GetFileName(name,time=-1):
	if name[-4:] == ".txt":
		return name
	else:
		tExtension = ''
		if time>-1:
			tExtension = '_t{t:05d}'.format(t=time)
		return name + tExtension + '.txt'

def GetPolylineName(name,pl=-1,time=-1):
	if name[-4:] == ".txt":
		return name
	else:
		pExtension = ''
		tExtension = ''
		if pl>-1:
			pExtension = '{p:02d}'.format(p=pl)
		if time>-1:
			tExtension = '_t{t:05d}'.format(t=time)
		return name + pExtension + tExtension + '.txt'

def ImportPolyline(where='.',name='tpcl.txt',skiprows=0,info=True):
	my_file = where + '/' + name
	xs,ys,zs = np.loadtxt(my_file,usecols=(0,1,2),skiprows=skiprows,unpack=True)
	if (xs.size == 0):
		print("Error: empty file")
		sys.exit()
	if info:
		print("  Imported {} points from {}/{}".format(xs.size,where,name))
	return np.transpose([xs,ys,zs])

def GetPolylineLength(polyline):
	if polyline.size > 0:
		diff = np.diff(polyline,axis=0)
		norm = np.linalg.norm(diff,axis=1)
		return np.sum(norm)
	return 0

def ImportPolylines(where='.',name='tpcl',cloudsize=-1,time=-1,smooth=-1,info=True):
	fs = []
	polylines = []
	if cloudsize > 0:
		#cloud of polylines
		for pl in range(cloudsize):
			f = GetPolylineName(name=name,pl=pl,time=time)
			fs.append(f)
	else:
		#single polyline
		f = GetPolylineName(name=name,time=time)
		fs.append(f)
	for f in fs:
		if exists(where + '/' + f):
			polyline = ImportPolyline(where=where,name=f,info=info)
			if (smooth > 0):
				polyline = SmoothPolyline(polyline,smooth=smooth,info=True)
			polylines.append(polyline)
		else:
			if info:
				print('  Warning: {} does not exist'.format(where+'/'+f))
			polylines.append(np.array([]))
	return polylines

def GetSide(where='.',name='side0',time=-1,transform=True):
	fname = GetFileName(where+'/'+name,time)
	data = np.loadtxt(fname,usecols=0,unpack=True)
	m = np.reshape(data[0:9],(3,3)).transpose()
	h = data[9]
	n = data[10:13]
	c = data[13:16]
	q = Quaternion(data[19],data[16],data[17],data[18])
	if transform:
		m = np.array([c + q.rotate(pt) for pt in m])
		n = q.rotate(n)
	return m[0],m[1],m[2],h,n,c,q

#def GetHinge(where='.',name='hinge0',time):
	#hinge_name = GetFileName(where+'/hinge0',time)
	#r,x,y,z,qx,qy,qz,qw = np.loadtxt(fname,usecols=(1,4,5,6,7,8,9,10),unpack=True)
	##length = (r.size-1) * (2*r[0])
	##rs = r / length
	##pts = np.stack([x,y,z],axis=-1)
	##pts = (pts - pts[0])/ length
	##qs = np.stack((qw,qx,qy,qz),axis=-1)
	##qs = [Quaternion(q) for q in qs]
	##return rs[0], pts, qs

def GetReferenceForce(where='.',name='hinge0',time=0):
	hinge_name = GetFileName(where+'/hinge0',time)
	radii = np.loadtxt(hinge_name,usecols=(1))
	radius = radii[0]
	n = radii.size
	ks = 1
	length = 2*(n-1)*radius
	return 0.5 * ks * (radius**3) / (length**2)

def GetCapillaryForce(where='.',name='tpcl',theta=90.,time=-1,info=True):
	if info:
		print('  path = {}'.format(where))
		print('  theta = {}'.format(theta))
	sinTheta = np.sin(np.pi*theta/180)
	def GetPivot(where,time):
		A,B,C,_,_,_,_ = GetSide(where=where,time=time)
		return (A+B)/2
	def GetVDir(where,time):
		A,B,C,_,n,_,_ = GetSide(where=where,time=time)
		return n
	def GetForce(polylines,pivot,n):
		forces,arms = [],[]
		for polyline in polylines:
			length = GetPolylineLength(polyline)
			if length>0:
				force = length * n
				arm = np.mean(polyline,axis=0) - pivot
				forces.append(force)
			else:
				forces.append(np.zeros(3))
				arms.append(np.zeros(3))
		forces = np.array(forces) * sinTheta
		torques = np.array([np.cross(arm,force) for arm,force in zip(arms,forces)])
		return np.sum(forces,axis=0),np.sum(torques,axis=0)
	polylines = ImportPolylines(where = where,
	                            name = name,
	                            cloudsize = 10, #extra large
	                            time = time,
	                            smooth = -1,
	                            info = info)
	pivot = GetPivot(where,time)
	n = GetVDir(where,time)
	Fc,Tc = GetForce(polylines,pivot,n)
	return Fc,Tc

def GetRealCapillaryForce(where='.',name='tpcl',theta=90,time=-1,ec=1,vb=3700,info=True):
	if info:
		print('  ec = {}'.format(ec))
	F,T = GetCapillaryForce(where=where,name=name,time=time,theta=theta,info=info)
	F0 = GetReferenceForce(where=where,time=time)
	L0_pow3 = vb/(2*np.pi/3)
	L0 = L0_pow3**(1/3)
	L0_pow2 = L0_pow3**(2/3)
	F = ec*F*F0/L0
	T = ec*T*F0/L0_pow2
	return F,T

def ParaviewScript(where='.',sides=4,time=399,script='plot3D.py'):
	time_extension=''
	if time>0:
		time_extension='_t{:05}'.format(time)
	original_stdout = sys.stdout
	with open(script,'a') as f:
		sys.stdout = f
		print('from paraview.simple import *')
		print('base = STLReader(registrationName = \'base\', FileNames = [\'{pth}/base.stl\'])'.format(pth=where))
		for side in range(sides):
			print('side{side} = STLReader(registrationName = \'side{side}\', FileNames = [\'{pth}/side{side}{t_ext}.stl\'])'.format(pth=where,side=side,t_ext=time_extension))
		print('drop = LegacyVTKReader(registrationName=\'drop\', FileNames=[\'{pth}/drop{t_ext}.vtk\'])'.format(pth=where,t_ext=time_extension))
		print('renderView = GetActiveViewOrCreate(\'RenderView\')')
		print('renderView.OrientationAxesVisibility = 0')
		print('base_display = Show(base, renderView, \'GeometryRepresentation\')')
		print('ColorBy(base_display, None)')
		for side in range(sides):
			print('side{side}_display = Show(side{side}, renderView, \'GeometryRepresentation\')'.format(side=side))
			print('ColorBy(side{side}_display, None)'.format(side=side))
		print('drop_display = Show(drop, renderView, \'GeometryRepresentation\')')
		print('ColorBy(drop_display, None)')
		print('drop_display.AmbientColor = [0.0,0.666,1.0]')
		print('drop_display.DiffuseColor = [0.0,0.666,1.0]')
		print('drop_display.Opacity = 0.8')
	sys.stdout = original_stdout
	return 0
